/*
 * SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
 *
 * SPDX-License-Identifier: AGPL-3.0-only
 */

div.body {
    max-width: 90% !important;
    font-size: 11pt;
}

div.sphinxsidebar {
    font-size: 11pt;
}
